/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

// TODO allocation needs matching deletes

#include "surface.pmh"

#define FE_QT_VERBOSE			FALSE
#define FE_QT_DEBUG				FALSE
#define FE_QT_DRAW_NORMALS		FALSE
#define FE_QT_DRAW_SPHERES		FALSE
#define FE_QT_DRAW_SUBDIV		TRUE
#define FE_QT_DRAW_TEXT			FALSE
#define FE_QT_CLUMP_NEARBY		FALSE

#define FE_QT_SUBDIV_AGE		TRUE
#define FE_QT_SUBDIV_AGE_MAX	0

#define FE_QT_CHECK				(FE_CODEGEN<=FE_DEBUG)

namespace fe
{

QuadTree::QuadTree(void):
	m_divisions(2,2,1),
	m_capacity(4),
	m_subdivMax(2),
	m_ageStamp(0),
	m_timeStamp(0)
{
	feLog("QuadTree::QuadTree\n");

	m_spRootNode=new Node;

	m_spWireframe=new DrawMode;
	m_spWireframe->setDrawStyle(DrawMode::e_outline);

	String value;
	if(System::getEnvironmentVariable("FE_SUBDIV", value))
	{
		m_subdivMax=atoi(value.c_str());
	}
}

void QuadTree::populate(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,U32 a_vertices,
	const SpatialVector& a_center,F32 a_radius)
{
	feLog("QuadTree::populate\n");

	//* TODO cache a copy?
	m_pVertexArray=a_pVertexArray;
	m_pNormalArray=a_pNormalArray;
	m_vertices=a_vertices;
	m_sphereCount=m_vertices/m_grain;

	cacheSpheres(a_center,a_radius);
	buildGraph();
}

void QuadTree::cacheSpheres(const SpatialVector& a_center,F32 a_radius)
{
	Element* pElements=new Element[m_sphereCount];
	m_pSphereArray=(Vector<4,Real>*)reallocate(m_pSphereArray,
			m_sphereCount*sizeof(Vector<4,Real>));
	const F32 radius2=a_radius*a_radius;

#if FE_QT_DEBUG
		feLog("QuadTree::cacheSpheres count %d radius %.6G\n",
				m_sphereCount,a_radius);
#endif
	U32 elements=0;
	for(U32 m=0;m<m_sphereCount;m++)
	{
		SpatialVector center;
		Real radius;
		const U32 m3=m*3;
		BoundingSphere<Real>::solve(&m_pVertexArray[m3],3,center,radius);

#if FE_QT_DEBUG
		feLog("sphere %2d element %2d center %s radius %.6G (%.6G vs %.6G)\n",
				m,elements,
				c_print(center),radius,
				magnitudeSquared(center),radius2);
#endif

		if(magnitudeSquared(center-a_center)>radius2)
		{
			continue;
		}

		m_pSphereArray[elements]=center;
		m_pSphereArray[elements][3]=radius;

		pElements[elements].m_index=m;
		pElements[elements].m_sphereIndex=elements;
		elements++;

#if FALSE
		for(U32 n=0;n<3;n++)
		{
			feLog("  %s\n",c_print(m_pVertexArray[m*3+n]));
		}
		feLog("QuadTree::cacheSpheres %d %s\n",
				m,c_print(m_pSphereArray[m]));
#endif
	}
	m_sphereCount=elements;
	m_pSphereArray=(Vector<4,Real>*)reallocate(m_pSphereArray,
			m_sphereCount*sizeof(Vector<4,Real>));

	m_spRootNode->m_elementCount=m_sphereCount;
	m_spRootNode->m_pElements=new Element[m_sphereCount];
	for(U32 m=0;m<m_sphereCount;m++)
	{
		m_spRootNode->m_pElements[m]=pElements[m];
	}
	delete[] pElements;
}

// TODO deletes
void QuadTree::buildGraph(void)
{
#if FE_QT_DEBUG
	if(m_vertices<10000)
	{
		for(U32 m=0;m<m_vertices;m++)
		{
			feLog("vertex %2d %s  %s\n",m,c_print(m_pVertexArray[m]),
					c_print(m_pNormalArray[m]));
		}
	}
#endif

#if FE_QT_CHECK
	for(U32 m=0;m<m_vertices;m+=m_grain)
	{
		for(U32 n=0;n<m_grain-1;n++)
		{
			for(U32 p=n+1;p<m_grain-1;p++)
			{
				if(magnitudeSquared(m_pVertexArray[m+n]-
						m_pVertexArray[m+p])<1e-6)
				{
					feLog("QuadTree::buildGraph"
							" triangle with coincident vertices"
							" %d %d  %s  %s\n",
							m+n,m+p,c_print(m_pVertexArray[m+n]),
							c_print(m_pVertexArray[m+p]));
				}
			}
		}
	}
#endif

	m_spRootNode->calcBounds(m_pVertexArray,m_grain);

	SpatialVector* pVertexScratch=
			new SpatialVector[((m_spRootNode->m_elementCount*m_grain)<<1)+1];

	m_spRootNode->refine(m_pVertexArray,m_pNormalArray,pVertexScratch,m_grain,
			m_divisions,m_pSphereArray,m_capacity);

	delete[] pVertexScratch;
}

void QuadTree::Node::calcBounds(const SpatialVector* a_pVertexArray,U32 a_grain)
{
	SpatialVector center;
	Real radius;
	BoundingSphere<Real>::solve(a_pVertexArray,m_elementCount*a_grain,
			center,radius);
	m_boundSphere=center;
	m_boundSphere[3]=radius;

#if FE_QT_DEBUG
	feLog("QuadTree::Node::calcBounds %s\n",c_print(m_boundSphere));
#endif
}

void QuadTree::Node::calcBounds(const SpatialVector* a_pVertexArray,
	SpatialVector* a_pVertexScratch,U32 a_grain)
{
	for(U32 m=0;m<m_elementCount;m++)
	{
		const U32 index=m_pElements[m].m_index;
		for(U32 n=0;n<a_grain;n++)
		{
			a_pVertexScratch[m*a_grain+n]=a_pVertexArray[index*a_grain+n];
		}
	}

	calcBounds(a_pVertexScratch,a_grain);
}

void QuadTree::Node::refine(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,
	SpatialVector* a_pVertexScratch, U32 a_grain,const Vector3u& a_divisions,
	const Vector4* a_pSphereArray,U32 a_capacity)
{
#if FE_QT_DEBUG
	feLog("refine %p %d\n",this,m_elementCount);
	for(U32 m=0;m<m_elementCount;m++)
	{
		const U32 index=m_pElements[m].m_index;
		const U32 sphereIndex=m_pElements[m].m_sphereIndex;
		feLog(" %d:%.3f %d",
				index,m_pElements[m].m_sortValue,sphereIndex);
		const Vector4& rVector=a_pSphereArray[sphereIndex];
		feLog(" (%.3f,%.3f)",
				rVector[0],rVector[1]);
	}
	feLog("\n");
#endif

	if(m_elementCount<=a_capacity)
	{
#if FALSE
		//* TEMP testing
		const U32 maxLevel=4;
		subdivideAll(a_pVertexArray,a_pNormalArray,a_grain,1,maxLevel);
#endif

		return;
	}

	FEASSERT(a_divisions[0]>0);
	FEASSERT(a_divisions[0]<3);	// TEMP
	FEASSERT(a_divisions[1]>0);
	FEASSERT(a_divisions[1]<3);	// TEMP
	FEASSERT(a_divisions[2]>0);
	FEASSERT(a_divisions[2]<3);	// TEMP

	//* `X'
	const U32 divisionsX=a_divisions[0];
	Element** ppElementsX=new Element*[divisionsX];
	U32* pElementCountX=new U32[divisionsX];
	if(divisionsX>1)
	{
		split(a_pSphereArray,0,m_pElements,m_elementCount,
				divisionsX,ppElementsX,pElementCountX);
	}
	else
	{
		ppElementsX[0]=m_pElements;
		pElementCountX[0]=m_elementCount;
	}

#if FE_QT_DEBUG
	for(U32 n=0;n<divisionsX;n++)
	{
		feLog("refine X %d/%d %p %d\n",n,divisionsX,this,pElementCountX[n]);
		for(U32 m=0;m<pElementCountX[n];m++)
		{
			const U32 index=ppElementsX[n][m].m_index;
			const U32 sphereIndex=ppElementsX[n][m].m_sphereIndex;
			feLog(" %d:%.3f %d",
					index,ppElementsX[n][m].m_sortValue,sphereIndex);
			const Vector4& rVector=a_pSphereArray[sphereIndex];
			feLog(" (%.3f,%.3f)",
					rVector[0],rVector[1]);
		}
	feLog("\n");
	}
#endif

	//* `Y'
	const U32 divisionsY=a_divisions[1];
	const U32 totalY=divisionsX*divisionsY;
	Element** ppElementsY=new Element*[totalY];
	U32* pElementCountY=new U32[totalY];
	if(divisionsY>1)
	{
		for(U32 m=0;m<divisionsX;m++)
		{
			split(a_pSphereArray,1,ppElementsX[m],pElementCountX[m],
					divisionsY,&ppElementsY[m*divisionsY],
					&pElementCountY[m*divisionsY]);
		}
	}
	else
	{
		for(U32 m=0;m<divisionsX;m++)
		{
			ppElementsY[m]=ppElementsX[m];
			pElementCountY[m]=pElementCountX[m];
		}
	}

	delete[] pElementCountX;
	delete[] ppElementsX;

#if FE_QT_DEBUG
	for(U32 n=0;n<divisionsY;n++)
	{
		feLog("refine Y %d/%d %p %d\n",n,divisionsY,this,pElementCountY[n]);
		for(U32 m=0;m<pElementCountY[n];m++)
		{
			const U32 index=ppElementsY[n][m].m_index;
			const U32 sphereIndex=ppElementsY[n][m].m_sphereIndex;
			feLog(" %d:%.3f %d",
					index,ppElementsY[n][m].m_sortValue,sphereIndex);
			const Vector4& rVector=a_pSphereArray[sphereIndex];
			feLog(" (%.3f,%.3f)",
					rVector[0],rVector[1]);
		}
	feLog("\n");
	}
#endif

	//* `Z'
	const U32 divisionsZ=a_divisions[2];
	const U32 totalZ=divisionsX*divisionsY*divisionsZ;
	Element** ppElementsZ=new Element*[totalZ];
	U32* pElementCountZ=new U32[totalZ];
	if(divisionsZ>1)
	{
		for(U32 m=0;m<totalY;m++)
		{
			split(a_pSphereArray,2,ppElementsY[m],pElementCountY[m],
					divisionsZ,&ppElementsZ[m*divisionsZ],
					&pElementCountZ[m*divisionsZ]);
		}
	}
	else
	{
		for(U32 m=0;m<totalY;m++)
		{
			ppElementsZ[m]=ppElementsY[m];
			pElementCountZ[m]=pElementCountY[m];
		}
	}

	delete[] pElementCountY;
	delete[] ppElementsY;

#if FE_QT_DEBUG
	for(U32 n=0;n<divisionsZ;n++)
	{
		feLog("refine Z %d/%d %p %d\n",n,divisionsZ,this,pElementCountZ[n]);
		for(U32 m=0;m<pElementCountZ[n];m++)
		{
			const U32 index=ppElementsZ[n][m].m_index;
			const U32 sphereIndex=ppElementsZ[n][m].m_sphereIndex;
			feLog(" %d:%.3f %d",
					index,ppElementsZ[n][m].m_sortValue,sphereIndex);
			const Vector4& rVector=a_pSphereArray[sphereIndex];
			feLog(" (%.3f,%.3f)",
					rVector[0],rVector[1]);
		}
	feLog("\n");
	}
#endif

	//* if all elements fall into one branch, stop splitting
	U32 loner=FALSE;
	for(U32 m=0;m<totalZ;m++)
	{
		if(pElementCountZ[m]==m_elementCount)
		{
			loner=TRUE;
			break;
		}
	}

	if(!loner)
	{
		for(U32 m=0;m<totalZ;m++)
		{
			if(pElementCountZ[m])
			{
				sp<Node> spChild(new Node);
				spChild->m_level=m_level+1;
				spChild->m_pElements=ppElementsZ[m];
				spChild->m_elementCount=pElementCountZ[m];
				spChild->attachTo(sp<Node>(this));
				spChild->calcBounds(a_pVertexArray,a_pVertexScratch,a_grain);
				spChild->refine(a_pVertexArray,a_pNormalArray,a_pVertexScratch,
						a_grain,a_divisions,a_pSphereArray,a_capacity);
			}
		}
	}

	delete[] pElementCountZ;
	delete[] ppElementsZ;
}

// TODO currently assumes divisions==2
void QuadTree::Node::split(const Vector4* a_pSphereArray,U32 axis,
	Element*& a_rpElements,U32& a_rElementCount,
	U32 divisions,Element** a_rppElements,U32* a_rpElementCount)
{
	FEASSERT(divisions==2);
#if FE_QT_DEBUG
	feLog("QuadTree::Node::split %d by %d\n",
			a_rElementCount,divisions);
#endif
	for(U32 m=0;m<a_rElementCount;m++)
	{
		a_rpElements[m].m_sortValue=
				a_pSphereArray[a_rpElements[m].m_sphereIndex][axis];
#if FE_QT_DEBUG
		feLog("QuadTree::Node::split element %d %s\n",
				m,c_print(a_pSphereArray[a_rpElements[m].m_sphereIndex]));
#endif
	}
	qsort(a_rpElements,a_rElementCount,sizeof(Element),&Element::compare);
	U32 n=a_rElementCount>>1;
#if FE_QT_CLUMP_NEARBY
	if(n)
	{
		//* try to keep a little of existing organization
		while(n<a_rElementCount &&
				a_rpElements[n].m_sortValue-a_rpElements[n-1].m_sortValue<1e-3)
		{
			n++;
		}
	}
#endif
#if FE_QT_DEBUG
	feLog("QuadTree::Node::split lvl %d: %d at %d axis %d\n",
			m_level,a_rElementCount,n,axis);
#endif
	a_rpElementCount[0]=n;
	a_rpElementCount[1]=a_rElementCount-a_rpElementCount[0];
	a_rppElements[0]=new Element[a_rpElementCount[0]];
	a_rppElements[1]=new Element[a_rpElementCount[1]];
	for(U32 m=0;m<a_rpElementCount[0];m++)
	{
		a_rppElements[0][m]=a_rpElements[m];
	}
	for(U32 m=0;m<a_rpElementCount[1];m++)
	{
		a_rppElements[1][m]=a_rpElements[a_rpElementCount[0]+m];
	}
	delete[] a_rpElements;
	a_rpElements=NULL;
	a_rElementCount=0;
}

void QuadTree::BaseNode::calcFrames(
	SpatialTransform& a_rTransform1,SpatialTransform& a_rTransform2,
	const SpatialVector& a_rVertex1,const SpatialVector& a_rNormal1,
	const SpatialVector& a_rVertex2,const SpatialVector& a_rNormal2)
{
	const SpatialVector direction=unit(a_rVertex2-a_rVertex1);
	a_rTransform1.left()=unit(cross(a_rNormal1,direction));
	a_rTransform2.left()=unit(cross(a_rNormal2,direction));
	a_rTransform1.direction()=cross(a_rTransform1.left(),a_rNormal1);
	a_rTransform2.direction()=cross(a_rTransform2.left(),a_rNormal2);
	a_rTransform1.up()=a_rNormal1;
	a_rTransform2.up()=a_rNormal2;
	a_rTransform1.translation()=a_rVertex1;
	a_rTransform2.translation()=a_rVertex2;

#if FALSE
	feLog("direction\n%s\n",c_print(direction));
	feLog("frame1\n%s\n",c_print(a_rTransform1));
	feLog("frame2\n%s\n",c_print(a_rTransform2));
#endif
}

void QuadTree::Node::subdivideOne(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,
	U32 a_triIndex,U32 a_grain,U32 a_level,U32 a_maxLevel,U32 a_timeStamp)
{
	subdivide(a_pVertexArray,a_pNormalArray,&m_pElements[a_triIndex],
			a_triIndex,1,a_grain,a_level,a_maxLevel,a_timeStamp);
}

void QuadTree::Node::subdivideAll(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,U32 a_grain,U32 a_level,
	U32 a_maxLevel,U32 a_timeStamp)
{
	subdivide(a_pVertexArray,a_pNormalArray,m_pElements,0,m_elementCount,
			a_grain,a_level,a_maxLevel,a_timeStamp);
}

void QuadTree::SubNode::subdivideOne(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,
	U32 a_triIndex,U32 a_grain,U32 a_level,U32 a_maxLevel,U32 a_timeStamp)
{
	Element pElements[1];
	pElements[0].m_index=a_triIndex;
	subdivide(m_pSubdivVertexArray,m_pSubdivNormalArray,
			pElements,a_triIndex,1,a_grain,a_level,a_maxLevel,a_timeStamp);
}

void QuadTree::SubNode::subdivideAll(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,U32 a_grain,U32 a_level,
	U32 a_maxLevel,U32 a_timeStamp)
{
	U32 elementCount=triangleCount();
	Element* pElements=new Element[elementCount];
	for(U32 m=0;m<elementCount;m++)
	{
		pElements[m].m_index=m;
	}
	subdivide(m_pSubdivVertexArray,
			m_pSubdivNormalArray,
			pElements,0,elementCount,
			a_grain,a_level,a_maxLevel,a_timeStamp);
	delete[] pElements;
}

void QuadTree::BaseNode::subdivide(const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,
	const Element* a_pElements,U32 a_elementIndex,U32 a_elementCount,
	U32 a_grain,U32 a_level,U32 a_maxLevel,U32 a_timeStamp)
{
#if FE_QT_DEBUG
	feLog("QuadTree::BaseNode::subdivide"
			" a_grain=%d a_elementIndex=%d a_elementCount=%d a_level=%d\n",
			a_grain,a_elementIndex,a_elementCount,a_level);
#endif

	for(U32 m=0;m<a_elementCount;m++)
	{
		const U32 serial=a_elementIndex+m;

		//* check for preexisting matching sibling
//		BaseNode* pParent=dynamic_cast<QuadTree::BaseNode*>(*parent());
//		FEASSERT(pParent);
//		List<DAGNode*>::Iterator it=
//				const_cast<QuadTree::BaseNode*>(pParent)->child();
		List< sp<DAGNode> >::Iterator it=
				const_cast<QuadTree::BaseNode*>(this)->child();
		sp<DAGNode> spSibling;
		BWORD found=FALSE;
		while((spSibling= *it++).isValid())
		{
			sp<QuadTree::SubNode> spSubNode=spSibling;
#if FE_QT_DEBUG
			if(spSubNode.isValid())
			{
				feLog("  existing %d\n",spSubNode->m_serial);
			}
#endif
			if(spSubNode.isValid() && spSubNode->m_serial==serial)
			{
				spSubNode->m_timeStamp=a_timeStamp;
				found=TRUE;
				break;
			}
		}
		if(found)
		{
#if FE_QT_DEBUG
			feLog("QuadTree::BaseNode::subdivide found %d\n",serial);
#endif
			continue;
		}

		sp<SubNode> spChild(new SubNode);
		spChild->attachTo(sp<DAGNode>(this));
		spChild->m_level=m_level+1;
		spChild->m_timeStamp=a_timeStamp;
		spChild->m_serial=serial;
		spChild->m_subdivCount=12;
		spChild->m_subdivLevel=a_level;
		spChild->m_pSubdivVertexArray=new SpatialVector[spChild->m_subdivCount];
		spChild->m_pSubdivNormalArray=new SpatialVector[spChild->m_subdivCount];

		const U32 offset=a_grain*a_pElements[m].m_index;

		const SpatialVector& v0=a_pVertexArray[offset];
		const SpatialVector& v1=a_pVertexArray[offset+1];
		const SpatialVector& v2=a_pVertexArray[offset+2];

		const SpatialVector& n0=a_pNormalArray[offset];
		const SpatialVector& n1=a_pNormalArray[offset+1];
		const SpatialVector& n2=a_pNormalArray[offset+2];

		if(isZero(v1-v0) || isZero(v2-v0) || isZero(v2-v1))
		{
			continue;
		}

#if FE_QT_DEBUG
		printf("element %d/%d index %d offset %d\n",m,a_elementCount,
				a_pElements[m].m_index,offset);
		printf("v0 %s  %s\n",c_print(v0),c_print(n0));
		printf("v1 %s  %s\n",c_print(v1),c_print(n1));
		printf("v2 %s  %s\n",c_print(v2),c_print(n2));
#endif

#if FALSE
		// linear
		const SpatialVector mid01=0.5*(v0+v1);
		const SpatialVector mid12=0.5*(v1+v2);
		const SpatialVector mid20=0.5*(v2+v0);
#endif

#if FALSE
		// MatrixSqrt
		MatrixSqrt<SpatialTransform> matrixSqrt;
		SpatialTransform frame0;
		SpatialTransform frame1;
		calcFrames(frame0,frame1,v0,n0,v1,n1);
		SpatialTransform inv0;
		invert(inv0,frame0);
		SpatialTransform delta=inv0*frame1;
		SpatialTransform partial;
		matrixSqrt.solve(partial,delta);
		SpatialTransform mid=frame0*partial;
		SpatialVector mid01=mid.translation();
		SpatialVector norm01=mid.up();

		FEASSERT(magnitude(norm01)>0.99);
		FEASSERT(magnitude(norm01)<1.01);

		calcFrames(frame0,frame1,v1,n1,v2,n2);
		invert(inv0,frame0);
		delta=inv0*frame1;
		matrixSqrt.solve(partial,delta);
		mid=frame0*partial;
		SpatialVector mid12=mid.translation();
		SpatialVector norm12=mid.up();

		FEASSERT(magnitude(norm12)>0.99);
		FEASSERT(magnitude(norm12)<1.01);

		calcFrames(frame0,frame1,v2,n2,v0,n0);
		invert(inv0,frame0);
		delta=inv0*frame1;
		matrixSqrt.solve(partial,delta);
		mid=frame0*partial;
		SpatialVector mid20=mid.translation();
		SpatialVector norm20=mid.up();

		FEASSERT(magnitude(norm20)>0.99);
		FEASSERT(magnitude(norm20)<1.01);
#else
		TrianglePN<Real> trianglePN;
		trianglePN.configure(v0,v1,v2,n0,n1,n2);

		const SpatialVector mid01=
				trianglePN.midpoint(TrianglePN<Real>::e_v1v2);
		const SpatialVector mid12=
				trianglePN.midpoint(TrianglePN<Real>::e_v2v3);
		const SpatialVector mid20=
				trianglePN.midpoint(TrianglePN<Real>::e_v3v1);

		const SpatialVector norm01=
				trianglePN.midnormal(TrianglePN<Real>::e_v1v2);
		const SpatialVector norm12=
				trianglePN.midnormal(TrianglePN<Real>::e_v2v3);
		const SpatialVector norm20=
				trianglePN.midnormal(TrianglePN<Real>::e_v3v1);
#endif

		spChild->m_pSubdivVertexArray[0]=v0;
		spChild->m_pSubdivVertexArray[1]=mid01;
		spChild->m_pSubdivVertexArray[2]=mid20;

		spChild->m_pSubdivVertexArray[3]=mid01;
		spChild->m_pSubdivVertexArray[4]=mid12;
		spChild->m_pSubdivVertexArray[5]=mid20;

		spChild->m_pSubdivVertexArray[6]=mid01;
		spChild->m_pSubdivVertexArray[7]=v1;
		spChild->m_pSubdivVertexArray[8]=mid12;

		spChild->m_pSubdivVertexArray[9]=mid12;
		spChild->m_pSubdivVertexArray[10]=v2;
		spChild->m_pSubdivVertexArray[11]=mid20;

		spChild->m_pSubdivNormalArray[0]=n0;
		spChild->m_pSubdivNormalArray[1]=norm01;
		spChild->m_pSubdivNormalArray[2]=norm20;

		spChild->m_pSubdivNormalArray[3]=norm01;
		spChild->m_pSubdivNormalArray[4]=norm12;
		spChild->m_pSubdivNormalArray[5]=norm20;

		spChild->m_pSubdivNormalArray[6]=norm01;
		spChild->m_pSubdivNormalArray[7]=n1;
		spChild->m_pSubdivNormalArray[8]=norm12;

		spChild->m_pSubdivNormalArray[9]=norm12;
		spChild->m_pSubdivNormalArray[10]=n2;
		spChild->m_pSubdivNormalArray[11]=norm20;

		SpatialVector center;
		Real radius;
		BoundingSphere<Real>::solve(spChild->m_pSubdivVertexArray,
				spChild->m_subdivCount,center,radius);
		spChild->m_boundSphere=center;
		spChild->m_boundSphere[3]=radius;

#if FE_QT_DEBUG
		feLog("QuadTree::BaseNode::subdivide %s\n",
				c_print(spChild->m_boundSphere));
#endif

		if(a_level<a_maxLevel)
		{
			U32 elementCount=4;
			Element* pElements=new Element[elementCount];
			for(U32 n=0;n<elementCount;n++)
			{
				pElements[n].m_index=m*12+n;
			}
			spChild->subdivide(spChild->m_pSubdivVertexArray,
					spChild->m_pSubdivNormalArray,
					pElements,0,elementCount,
					a_grain,a_level+1,a_maxLevel,a_timeStamp);
			delete[] pElements;
		}
	}
}

int QuadTree::Element::compare(const void* a_pVoid1,const void* a_pVoid2)
{
	const Element* pElement1=static_cast<const Element*>(a_pVoid1);
	const Element* pElement2=static_cast<const Element*>(a_pVoid2);
	FEASSERT(pElement1);
	FEASSERT(pElement2);
	FEASSERT(pElement1->m_index!=pElement2->m_index);
	return (pElement1->m_sortValue==pElement2->m_sortValue)? 0:
			(pElement1->m_sortValue>pElement2->m_sortValue)? 1: -1;
}

int QuadTree::Child::compare(const void* a_pVoid1,const void* a_pVoid2)
{
	const Child* pChild1=static_cast<const Child*>(a_pVoid1);
	const Child* pChild2=static_cast<const Child*>(a_pVoid2);
	FEASSERT(pChild1);
	FEASSERT(pChild2);
	return (pChild1->m_distance==pChild2->m_distance)? 0:
			(pChild1->m_distance>pChild2->m_distance)? 1: -1;
}

Real QuadTree::rayImpact(const SpatialVector& a_origin,
	const SpatialVector& a_direction,U32 a_hitLimit,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray) const
{
//	feLog("QuadTree::rayImpact\n");
	const_cast<QuadTree*>(this)->prune();

	const BWORD actMutable=FALSE;
	const BWORD directional=TRUE;
	const U32 subdivMax=0;
	safeLockShared();
	Real distance=nearestTo(actMutable,directional,
			a_origin,a_direction,a_hitLimit,a_rHitArray,subdivMax);
	safeUnlockShared();
	return distance;
}

Real QuadTree::nearestPoint(const SpatialVector& a_origin,
	U32 a_hitLimit,std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray) const
{
	const_cast<QuadTree*>(this)->prune();

	const BWORD actMutable=TRUE;
	const BWORD directional=FALSE;
	const SpatialVector direction(0.0,0.0,0.0);
	safeLockShared();
	Real distance=nearestTo(actMutable,directional,
			a_origin,direction,a_hitLimit,a_rHitArray,m_subdivMax);
	safeUnlockShared();
	return distance;
}

Real QuadTree::nearestTo(BWORD a_mutable,BWORD a_directional,
	const SpatialVector& a_origin,const SpatialVector& a_direction,
	U32 a_hitLimit,std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
	U32 a_subdivMax) const
{
	const_cast<QuadTree*>(this)->prune();

#if FE_QT_VERBOSE
	feLog("\nQuadTree::nearestTo point %s\n",c_print(a_origin));
	if(a_directional)
	{
		feLog("  along ray %s\n",c_print(a_direction));
	}
#endif

	safeLockShared();
	m_spRootNode->nearestTo(a_mutable,a_directional,a_origin,a_direction,
			a_hitLimit,a_rHitArray,
			m_hitPool,a_subdivMax,m_pVertexArray,m_pNormalArray,m_grain,
			m_pSphereArray,m_timeStamp);
	safeUnlockShared();

#if FE_QT_VERBOSE
	const U32 hitCount=a_rHitArray.size();
	for(U32 m=0;m<hitCount;m++)
	{
		sp<SpatialTreeI::Hit>& rspHit=a_rHitArray[m];
		feLog("QuadTree::nearestTo hit %d/%d dist %.6G at %s\n",
				m,hitCount,rspHit->distance(),c_print(rspHit->intersection()));
		feLog("  bary %s\n",c_print(rspHit->barycenter()));
	}
#endif

	return a_rHitArray.size()? a_rHitArray[0]->distance(): 0.0;
}

Real QuadTree::BaseNode::separation(const SpatialVector& a_origin) const
{
	SpatialVector direction;
	const SpatialVector center=m_boundSphere;
	const Real radius=m_boundSphere[3];
	Real distance=
			PointSphereNearest<Real>::solve(center,radius,a_origin,direction);

	// balance disparate sizes: outside sphere is absolute; inside is a ratio
	if(distance<0.0)
	{
		distance/=radius;
	}
	return distance;
}

void QuadTree::BaseNode::nearestTo(BWORD a_mutable,BWORD a_directional,
	const SpatialVector& a_origin,const SpatialVector& a_direction,
	U32 a_hitLimit,std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray,
	CountedPool<Hit>& a_rHitPool,
	U32 a_subdivMax,const SpatialVector* a_pVertexArray,
	const SpatialVector* a_pNormalArray,U32 a_grain,
	const Vector4* a_pSphereArray,U32 a_timeStamp) const
{
	U32 hits=a_rHitArray.size();
	Real margin_distance=hits? a_rHitArray[hits-1]->distance(): -1.0;

	//* check sphere of entire node
	const Real gap=separation(a_origin);

#if FE_QT_DEBUG
	feLog("QuadTree::BaseNode::nearestTo %p level %d\n",this,m_level);
	if(!m_level)
	{
		feLog("\n");
	}
	feLog("nearestTo lvl %d tri %d child %d  %s sep %.6G vs %.6G\n",
			m_level,triangleCount(),childCount(),
			c_print(m_boundSphere),gap,margin_distance);
#endif

#if FE_QT_DRAW_TEXT
	if(a_mutable)
	{
		const_cast<QuadTree::BaseNode*>(this)->m_debugText.sPrintf("%.1f",gap);
	}
#endif

	if(margin_distance>=0.0 && gap>=margin_distance)
	{
#if FE_QT_DEBUG
		feLog("node reject\n");
#endif
		return;
	}

	U32 subdivChildren=0;
	List< sp<DAGNode> >::Iterator it=
			const_cast<QuadTree::BaseNode*>(this)->child();
	sp<DAGNode> spChild;
	while((spChild= *it++).isValid())
	{
		if(spChild.is<QuadTree::SubNode>())
		{
			subdivChildren++;
		}
	}

	// HACK const
	sp<BaseNode> spBaseNode(const_cast<QuadTree::BaseNode*>(this));

	sp<SubNode> spSubNode(spBaseNode);
	if(spSubNode.isValid() && a_mutable)
	{
		spSubNode->m_timeStamp=a_timeStamp;
	}

#if FE_QT_DRAW_TEXT
	if(a_mutable)
	{
		const_cast<QuadTree::BaseNode*>(this)->m_debugText.sPrintf("%.1fp",gap);
	}
#endif

	const U32 subdivLevel=spSubNode.isValid()? spSubNode->m_subdivLevel: 0;
	BWORD wouldSubdivide=(a_mutable && subdivLevel<a_subdivMax);

	U32 children=childCount();
	const U32 triangles=triangleCount();
	SpatialVector direction=a_direction;

#if FE_QT_DRAW_TEXT
	Real local_best= -1.0;
#endif

	// descendents are presumably better representations of the same data
	if(!children || subdivChildren)
	{
		//* check each element
		for(U32 m=0;m<triangles;m++)
		{
//			feLog("QuadTree::BaseNode::nearestTo array %p index %d\n",
//					a_pSphereArray,m);
			const SpatialVector* pVertexArray=triangleVertex(m,a_pVertexArray);
			const SpatialVector* pNormalArray=triangleNormal(m,a_pNormalArray);
			const Vector4 sphereVector=sphereData(m,a_pSphereArray);
//			feLog("QuadTree::BaseNode::nearestTo sphere %s\n",
//					c_print(sphereVector));
			const Real radius=sphereVector[3];

#if FE_QT_DEBUG
			feLog("sphere %s  vert %s  %s  %s\n",
					c_print(sphereVector),
					c_print(pVertexArray[0]),
					c_print(pVertexArray[1]),
					c_print(pVertexArray[2]));
#endif

			if(radius>0.0 && margin_distance>=0.0)
			{
				const SpatialVector center=sphereVector;
				if(a_directional)
				{
					Real distance=RaySphereIntersect<Real>::solve(center,radius,
							a_origin-radius*a_direction,a_direction);

					if(distance<0.0f)
					{
#if FE_QT_DEBUG
						feLog("skip %d  %s\n",m,c_print(sphereVector));
#endif
						continue;
					}
					distance-=radius;
				}
				else
				{
					if(!PointSphereNearest<Real>::within(
							center,radius,a_origin,margin_distance))
					{
#if FE_QT_DEBUG
						feLog("skip %d  %s\n",m,c_print(sphereVector));
#endif
						continue;
					}
				}
			}

			Barycenter<Real> barycenter;
			SpatialVector intersection;
			Real distance;
			if(a_directional)
			{
				distance=
					RayTriangleIntersect<Real>::solve(
					pVertexArray[0],
					pVertexArray[1],
					pVertexArray[2],
					a_origin,direction,
					barycenter);
				intersection=a_origin+direction*distance;
			}
			else
			{
				distance=
					PointTriangleNearest<Real>::solve(
					pVertexArray[0],
					pVertexArray[1],
					pVertexArray[2],
					a_origin,direction,
					intersection,barycenter);
			}

#if FE_QT_DEBUG
			feLog("tri  %d  %s  %.6G vs %.6G\n",
					m,c_print(sphereVector),distance,margin_distance);
#endif

#if FE_QT_DRAW_TEXT
			if(a_mutable && distance>=0 &&
					(local_best<0 || distance<local_best))
			{
				local_best=distance;
				const_cast<QuadTree::BaseNode*>(this)->m_debugText.sPrintf(
						"%.1f %.1f",gap, local_best);
			}
#endif

			if(wouldSubdivide)
			{
				// blend triangle distance with sphere separation
				const Real contraction=0.5;	//* TWEAK

				const Real contracted=contraction*distance+
						(1.0-contraction)*(gap<0.0? 0.0: gap);
				if(contracted>=0 &&
						(margin_distance<0 || contracted<margin_distance))
				{
					spBaseNode->subdivideOne(a_pVertexArray,a_pNormalArray,
							m,a_grain,subdivLevel+1,subdivLevel+1,a_timeStamp);
				}
				continue;
			}

			if(distance>=0 && (margin_distance<0 || distance<margin_distance))
			{
#if FE_QT_DEBUG
				feLog("BETTER\n");
				feLog("QuadTree::BaseNode::nearestTo %s  %s  %s\n",
						c_print(pVertexArray[0]),
						c_print(pVertexArray[1]),
						c_print(pVertexArray[2]));
#endif

				addHit(a_hitLimit,a_rHitArray,a_rHitPool,
						distance,direction,intersection,barycenter,
						pVertexArray[0],pVertexArray[1],pVertexArray[2],
						pNormalArray[0],pNormalArray[1],pNormalArray[2]);

				margin_distance=hits? a_rHitArray[hits-1]->distance(): -1.0;
			}
		}
	}

	children=childCount();
	if(children)
	{
		// TODO List should have a const Iterator
		Child* pChildOrder=new Child[children];
		sp<DAGNode> spChild;
		U32 checkCount=0;

		List< sp<DAGNode> >::Iterator it=
				const_cast<QuadTree::BaseNode*>(this)->child();
		while((spChild= *it++).isValid())
		{
			sp<SubNode> spSubNode(spChild);
			if(spSubNode.isValid())
			{
#if FE_QT_DEBUG
				feLog("nearestTo serial %d\n",spSubNode->m_serial);
#endif

//				spSubNode->m_timeStamp=a_timeStamp;
			}

			pChildOrder[checkCount].m_spNode=spChild;
			if(!pChildOrder[checkCount].m_spNode.isValid())
			{
#if FE_QT_DEBUG
				feLog("nearestTo not BaseNode\n");
#endif
				continue;
			}

			pChildOrder[checkCount].m_distance=
					pChildOrder[checkCount].m_spNode->separation(a_origin);
			checkCount++;
		}

		if(checkCount)
		{
			qsort(pChildOrder,checkCount,sizeof(Child),&Child::compare);
		}

		for(U32 m=0;m<checkCount;m++)
		{
#if FE_QT_DEBUG
			feLog("nearestTo lvl %d child %d/%d\n",m_level,m,checkCount);
#endif

			pChildOrder[m].m_spNode->nearestTo(a_mutable,a_directional,
					a_origin,a_direction,
					a_hitLimit,a_rHitArray,a_rHitPool,a_subdivMax,
					a_pVertexArray,a_pNormalArray,
					a_grain,a_pSphereArray,a_timeStamp);
		}
		delete[] pChildOrder;
	}
}

#if FALSE
void QuadTree::BaseNode::rayImpact(const SpatialVector& a_origin,
	const SpatialVector& a_direction,
	std::vector< sp<SpatialTreeI::Hit> >& a_rHitArray)
{
	Real best_distance= -1.0f;

/*
	for(U32 m=0;m<m_vertices;m+=m_grain)
	{
		const SpatialVector center=m_pSphereArray[m/m_grain];
		const Real radius=m_pSphereArray[m/m_grain][3];
		Real distance=RaySphereIntersect<Real>::solve(center,radius,
				a_origin-radius*a_direction,a_direction);
		if(distance<0.0f)
		{
			continue;
		}
		distance-=radius;

		Vector2 uv;
		distance=RayTriangleIntersect<Real>::solve(
				m_pVertexArray[m],
				m_pVertexArray[m+1],
				m_pVertexArray[m+2],
				a_origin,a_direction,uv[0],uv[1]);
		if(distance>=0.0 && (best_distance<0.0f ||
				distance<best_distance))
		{
			sp<Hit> spHit=m_hitPool.get();
			a_rHitArray.resize(1);
			a_rHitArray[0]=spHit;

			best_distance=distance;
			spHit->m_pVertex[0]=m_pVertexArray[m];
			spHit->m_pVertex[1]=m_pVertexArray[m+1];
			spHit->m_pVertex[2]=m_pVertexArray[m+2];
			spHit->m_pNormal[0]=m_pNormalArray[m];
			spHit->m_pNormal[1]=m_pNormalArray[m+1];
			spHit->m_pNormal[2]=m_pNormalArray[m+2];
			spHit->m_barycenter=uv;
			spHit->m_intersection=a_origin+distance*a_direction;
			spHit->m_direction=a_direction;
			spHit->m_distance=distance;
		}
	}
*/

	for(U32 m=0;m<m_vertices;m+=m_grain)
	{
		Vector2 uv;
		Real distance=RayTriangleIntersect<Real>::solve(
				m_pVertexArray[m],
				m_pVertexArray[m+1],
				m_pVertexArray[m+2],
				a_origin,a_direction,uv[0],uv[1]);
		if(distance>=0.0 && (best_distance<0.0f ||
				distance<best_distance))
		{
			sp<Hit> spHit=m_hitPool.get();
			a_rHitArray.resize(1);
			a_rHitArray[0]=spHit;

			best_distance=distance;
			spHit->m_pVertex[0]=m_pVertexArray[m];
			spHit->m_pVertex[1]=m_pVertexArray[m+1];
			spHit->m_pVertex[2]=m_pVertexArray[m+2];
			spHit->m_pNormal[0]=m_pNormalArray[m];
			spHit->m_pNormal[1]=m_pNormalArray[m+1];
			spHit->m_pNormal[2]=m_pNormalArray[m+2];
			spHit->m_barycenter=uv;
			spHit->m_intersection=a_origin+distance*a_direction;
			spHit->m_direction=a_direction;
			spHit->m_distance=distance;
		}
	}

	return best_distance;
}
#endif

void QuadTree::age(void)
{
	m_ageStamp++;
}

void QuadTree::prune(void)
{
#if FE_QT_DEBUG
	feLog("QuadTree::prune\n");
#endif
	if(m_timeStamp<m_ageStamp)
	{
		safeLock();
		//* if still less
		if(m_timeStamp<m_ageStamp)
		{
			m_spRootNode->prune(m_timeStamp);
			m_timeStamp=m_ageStamp;
		}
		safeUnlock();
	}
}

void QuadTree::BaseNode::prune(U32 a_timeStamp)
{
	const U32 maxAge=FE_QT_SUBDIV_AGE_MAX;

	const U32 children=childCount();
	if(children)
	{
		// TODO List should have a const Iterator
		List< sp<DAGNode> >::Iterator it=child();
		sp<DAGNode> spChild;
		while((spChild= *it++).isValid())
		{
			sp<QuadTree::BaseNode> spBaseNode(spChild);
			spBaseNode->prune(a_timeStamp);

			sp<SubNode> spSubNode(spChild);
#if FE_QT_DEBUG
			if(spSubNode.isValid())
			{
				feLog("QuadTree::BaseNode::prune %d %d  %s\n",
						spSubNode->m_timeStamp,a_timeStamp,
						c_print(spSubNode->m_boundSphere));
			}
#endif
			if(spSubNode.isValid() &&
					a_timeStamp-spSubNode->m_timeStamp>maxAge)
			{
#if FE_QT_DEBUG
				feLog("kill %p\n",spSubNode.raw());
#endif
				spSubNode->detach();
			}
		}
	}
}

void QuadTree::draw(sp<DrawI> a_spDrawI,const Color* a_pColor) const
{
#if FE_QT_SUBDIV_AGE
	// HACK wrong place
	QuadTree* pThis=const_cast<QuadTree*>(this);
	pThis->age();
#endif

	a_spDrawI->setDrawMode(m_spWireframe);

	safeLockShared();
	m_spRootNode->draw(a_spDrawI,m_pVertexArray,m_pNormalArray,
			m_grain,a_pColor,0);
	safeUnlockShared();

	a_spDrawI->setDrawMode(sp<DrawMode>(NULL));
}

void QuadTree::BaseNode::draw(sp<DrawI> a_spDrawI,
	const SpatialVector* a_pVertexArray,const SpatialVector* a_pNormalArray,
	U32 a_grain,const Color* a_pColor,U32 a_level) const
{
	Color white(1.0f,1.0f,1.0f,1.0f);
	Color red(1.0f,0.5f,0.5f,1.0f);
	Color autocolor(0.0f,0.0f,0.0f,1.0f);
	autocolor[a_level%3]=a_level<3? 0.5f: 1.0f;

	SpatialTransform transform;
	setIdentity(transform);
	translate(transform,m_boundSphere);
	const Real radius=m_boundSphere[3];
	const SpatialVector scale(radius,radius,radius);

#if FE_QT_DRAW_SPHERES
	a_spDrawI->drawSphere(transform,&scale,a_pColor? *a_pColor: autocolor);
#endif

#if FE_QT_DRAW_NORMALS || FE_QT_DRAW_TEXT
	const QuadTree::BaseNode* pBaseNode=
			dynamic_cast<const QuadTree::BaseNode*>(this);
	if(pBaseNode)
	{
		const Real normalScale=1.0;
		SpatialVector pLines[6];
		for(U32 m=0;m<pBaseNode->triangleCount();m++)
		{
			const SpatialVector* verts=
					pBaseNode->triangleVertex(m,a_pVertexArray);
			const SpatialVector* norms=
					pBaseNode->triangleNormal(m,a_pNormalArray);

			pLines[0]=verts[0];
			pLines[1]=verts[0]+norms[0]*normalScale;
			pLines[2]=verts[1];
			pLines[3]=verts[1]+norms[1]*normalScale;
			pLines[4]=verts[2];
			pLines[5]=verts[2]+norms[2]*normalScale;

			if(FE_QT_DRAW_NORMALS)
			{
				const SpatialVector* pNormals=NULL;
				const BWORD multicolor=FALSE;
				a_spDrawI->drawLines(pLines,pNormals,6,DrawI::e_discrete,
						multicolor,&red);
			}

			if(FE_QT_DRAW_TEXT)
			{
				const SpatialVector center=
						0.4*verts[0]+0.3*verts[1]+0.3*verts[2];
				a_spDrawI->drawAlignedText(center,m_debugText.c_str(),white);
			}
		}
	}
#endif

	const U32 children=childCount();
	if(children)
	{
		// TODO List should have a const Iterator
		List< sp<DAGNode> >::Iterator it=
				const_cast<QuadTree::BaseNode*>(this)->child();
		sp<DAGNode> spChild;
		while((spChild= *it++).isValid())
		{
#if FE_QT_DRAW_SUBDIV
			sp<SubNode> spSubNode(spChild);
			if(spSubNode.isValid() && spSubNode->m_subdivCount>2)
			{
//				feLog("draw %d %d\n",a_level,spSubNode->m_subdivCount);
//				for(U32 m=0;m<3;m++)
//				{
//					feLog("  %s\n",c_print(spSubNode->m_pSubdivVertexArray[m]));
//				}
				const Vector2f* pTexture=NULL;
				const BWORD multicolor=FALSE;
				a_spDrawI->drawTriangles(spSubNode->m_pSubdivVertexArray,
						spSubNode->m_pSubdivNormalArray,
						pTexture,spSubNode->m_subdivCount,
						DrawI::e_discrete,multicolor,
						a_pColor? a_pColor: &autocolor);
			}
#endif

			sp<QuadTree::BaseNode> spBaseNode(spChild);
			spBaseNode->draw(a_spDrawI,a_pVertexArray,a_pNormalArray,a_grain,
					a_pColor,a_level+1);
		}
	}
}

} // namespace
