/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_BLADE_DEBUG		FALSE

#include "grass/grass.pmh"

using namespace fe;

// NOTE dx and dv actually refer to angle and angular velocity,
// but we keep the familiar notation

Blade::Blade(void)
{
	//* current using uniform properties
	m_k=1e4;		// spring
	m_c=1e2;		// drag (angular)
	m_r=1e-2;		// rest
	m_m=1e0;		// mass
	m_h=1.0/60.0;	// time step

	populate();
}

void Blade::populate(void)
{
//	0	4	8	12
//	1	5	9	13
//	2	6	10	14
//	3	7	11	15

//	-2	1	0	0
//	1	-2	1	0   * k  (or c)
//	0	1	-2	1
//	0	0	1	-1

// non-symetrical?
//	-1	0	0	0
//	-1	-1	0	0   * restoration
//	-1	-1	-1	0
//	-1	-1	-1	-1

	set(m_dfdx);
	m_dfdx[0]= -2.0;
	m_dfdx[1]=1.0;

	m_dfdx[4]=1.0;
	m_dfdx[5]= -2.0;
	m_dfdx[6]=1.0;

	m_dfdx[9]=1.0;
	m_dfdx[10]= -2.0;
	m_dfdx[11]=1.0;

	m_dfdx[14]=1.0;
	m_dfdx[15]= -1.0;

	Matrix<4,4,Real> dfdv=m_dfdx;
	m_dfdx*=m_k;
	dfdv*=m_c;

	m_dfdx[0]-=m_restoration;
	m_dfdx[1]-=m_restoration;
	m_dfdx[2]-=m_restoration;
	m_dfdx[3]-=m_restoration;
	m_dfdx[5]-=m_restoration;
	m_dfdx[6]-=m_restoration;
	m_dfdx[7]-=m_restoration;
	m_dfdx[10]-=m_restoration;
	m_dfdx[11]-=m_restoration;
	m_dfdx[15]-=m_restoration;

	Matrix<4,4,Real> id;
	setIdentity(id);
	for(U32 m=0;m<4;m++)
	{
		id(m,m)*=4-m;
	}
	Matrix<4,4,Real> A=m_m*id-m_h*dfdv-m_h*m_h*m_dfdx;
	jad_invert(m_invA,A);

	Matrix<4,4,Real> check=A*m_invA;

#if FE_BLADE_DEBUG
	feLog("m_dfdx %s\n",c_print(m_dfdx));
	feLog("dfdv %s\n",c_print(dfdv));
	feLog("A %s\n",c_print(A));
	feLog("invA %s\n",c_print(m_invA));
	feLog("check %s\n",c_print(check));
#endif

	m_hdfdx=m_h*m_dfdx;
}

void Blade::step(U32 index,const std::vector<SpatialVector>& tangent,
	std::vector<SpatialVector>& location,
	std::vector<Real>& velocity) const
{
	const U32 index4=index*4;
	const U32 index5=index*5;

//	set(m_center,30.0*sin(m_t),0.0,30.0*cos(m_t));

	SpatialVector from;
	Vector<4,Real> push;	// external force
	set(push);

#if FALSE
	from=location[index5]-m_center;
	Real dist2=magnitudeSquared(from);
	if(dist2<m_radius*m_radius)
	{
		const Real alignment=dot(tangent[index],unit(from));
		setAll(push,5e2*(m_radius-sqrt(dist2))*alignment);
	}
#endif

	Vector<3,Real> d[4];	// delta segment
	d[0]=location[index5+1]-location[index5];
	d[1]=location[index5+2]-location[index5+1];
	d[2]=location[index5+3]-location[index5+2];
	d[3]=location[index5+4]-location[index5+3];

	Vector<4,Real> r;		// ground radial
	r[0]=sqrtf(d[0][0]*d[0][0]+d[0][2]*d[0][2]);
	r[1]=sqrtf(d[1][0]*d[1][0]+d[1][2]*d[1][2]);
	r[2]=sqrtf(d[2][0]*d[2][0]+d[2][2]*d[2][2]);
	r[3]=sqrtf(d[3][0]*d[3][0]+d[3][2]*d[3][2]);

	Vector<4,Real> l;		// length
	l[0]=sqrtf(d[0][0]*d[0][0]+d[0][1]*d[0][1]+d[0][2]*d[0][2]);
	l[1]=sqrtf(d[1][0]*d[1][0]+d[1][1]*d[1][1]+d[1][2]*d[1][2]);
	l[2]=sqrtf(d[2][0]*d[2][0]+d[2][1]*d[2][1]+d[2][2]*d[2][2]);
	l[3]=sqrtf(d[3][0]*d[3][0]+d[3][1]*d[3][1]+d[3][2]*d[3][2]);

	Vector<4,Real> a;		// angle
	for(U32 m=0;m<4;m++)
	{
		SpatialVector projected(d[m][0],0.0,d[m][2]);
		const Real projectedLength=magnitude(projected);
		Real tandot=1.0;
		if(projectedLength>0.0)
		{
			tandot=dot(tangent[index],
					unit(SpatialVector(d[m][0],0.0,d[m][2])));
		}
		a[m]=atan2f(tandot*r[m],d[m][1]);
	}

	Vector<4,Real> y;		// adjustment
	set(y);

	if(m_wind>0.0 || m_turbulence>0.0)
	{
		const Real locX=location[index5][0];
		const Real locZ=location[index5][2];
		const Real wind=tangent[index][0]*
				(m_wind+m_turbulence*sin(locX+m_t)*sin(locZ+m_t));
		push+=Vector<4,Real>(wind,wind,wind,wind);
	}

	push*=Vector<4,Real>(cos(a[0]),cos(a[1]),cos(a[2]),cos(a[3]));

	Vector<4,Real> w;		// angular velocity
	w[0]=velocity[index4];
	w[1]=velocity[index4+1];
	w[2]=velocity[index4+2];
	w[3]=velocity[index4+3];

	Vector<4,Real> f;		// force
	f[0]=-m_k*(a[0]-m_r)-m_c*(w[0])+
			m_k*(a[1]-a[0]-m_r)+m_c*(w[1]-w[0])-
			m_restoration*(a[0]-m_r);

	f[1]=-m_k*(a[1]-a[0]-m_r)-m_c*(w[1]-w[0])+
			m_k*(a[2]-a[1]-m_r)+m_c*(w[2]-w[1])-
			m_restoration*(a[1]-2.0*m_r);

	f[2]=-m_k*(a[2]-a[1]-m_r)-m_c*(w[2]-w[1])+
			m_k*(a[3]-a[2]-m_r)+m_c*(w[3]-w[2])-
			m_restoration*(a[2]-3.0*m_r);

	f[3]=-m_k*(a[3]-a[2]-m_r)-m_c*(w[3]-w[2])-
			m_restoration*(a[3]-4.0*m_r);
	f+=push;

	Vector<4,Real> b=m_h*(f+m_hdfdx*w+m_dfdx*y);
	Vector<4,Real> dv=m_invA*b;

#if FE_BLADE_DEBUG
	feLog("tangent %s\n",c_print(tangent[index]));
	feLog("push %s\n",c_print(push));
	feLog("f %s\n",c_print(f));
	feLog("b %s\n",c_print(b));
	feLog("dv %s\n",c_print(dv));
	feLog("l %s\n",c_print(l));
	feLog("a %s before\n",c_print(a));
	feLog("w %s\n",c_print(w));
#endif

	w+=dv;
	a+=m_h*w+y;

	//* TODO work this out properly
	const Real partX=tangent[index][0];
	const Real partZ=tangent[index][2];

	BWORD relocated=FALSE;
	for(U32 pass=0;pass<1+relocated;pass++)
	{
		const Real lso0=l[0]*sin(a[0]);
		location[index5+1][0]=location[index5][0]+lso0*partX;
		location[index5+1][1]=location[index5][1]+l[0]*cosf(a[0]);
		location[index5+1][2]=location[index5][2]+lso0*partZ;

		const Real lso1=l[1]*sin(a[1]);
		location[index5+2][0]=location[index5+1][0]+lso1*partX;
		location[index5+2][1]=location[index5+1][1]+l[1]*cosf(a[1]);
		location[index5+2][2]=location[index5+1][2]+lso1*partZ;

		const Real lso2=l[2]*sin(a[2]);
		location[index5+3][0]=location[index5+2][0]+lso2*partX;
		location[index5+3][1]=location[index5+2][1]+l[2]*cosf(a[2]);
		location[index5+3][2]=location[index5+2][2]+lso2*partZ;

		const Real lso3=l[3]*sin(a[3]);
		location[index5+4][0]=location[index5+3][0]+lso3*partX;
		location[index5+4][1]=location[index5+3][1]+l[3]*cosf(a[3]);
		location[index5+4][2]=location[index5+3][2]+lso3*partZ;

		if(pass)
		{
			break;
		}

		for(U32 m=0;m<4;m++)
		{
			const SpatialVector& base=location[index5+m];

			const SpatialVector toCenter=m_center-base;
			const Real reach=m_radius+l[m];
			if(magnitudeSquared(toCenter)<reach*reach)
			{

				//* point line-segment distance
				const SpatialVector span(sinf(a[m]),cosf(a[m]));
				Real along=dot(toCenter,span)/l[m];
				if(along<0.0f)
				{
					from=base-m_center;
				}
				else if(along>1.0f)
				{
					from=base+l[m]*span-m_center;
				}
				else
				{
					SpatialVector projection=base+along*l[m]*span;
					from=projection-m_center;
				}

				Real distance=magnitude(from);

#if FE_BLADE_DEBUG
				feLog("%d angle %.6G dist %.6G along %.6G\n",
						m,a[m],distance,along);
#endif

				if(distance>0.0f && distance<m_radius)
				{
					relocated=TRUE;

					SpatialVector offset=from*(m_radius-distance)/distance;

					d[m]=location[index5+1+m]-location[index5+m];

					SpatialVector d_mod=d[m]+offset;
					Real r_mod=sqrtf(d_mod[0]*d_mod[0]+d_mod[2]*d_mod[2]);
					Real tandot=dot(tangent[index],
							unit(SpatialVector(d_mod[0],0.0,d_mod[2])));
					Real a_mod=atan2f(tandot*r_mod,d_mod[1]);

					y[m]=a_mod-a[m];
					w[m]=0.0;

#if FE_BLADE_DEBUG
					feLog("  offset %s\n",c_print(offset));
					feLog("  d[m] %s\n",c_print(d[m]));
					feLog("  d_mod %s\n",c_print(d_mod));
					feLog("  r[m] %.6G\n",r[m]);
					feLog("  r_mod %.6G\n",r_mod);
					feLog("  angle %.6G y[m] %.6G\n",a_mod,y[m]);
#endif
				}
			}
			if(relocated)
			{
				const Real lsom=l[m]*sin(a[m]);
				location[index5+1+m][0]=location[index5+m][0]+lsom*partX;
				location[index5+1+m][1]=location[index5+m][1]+l[m]*cosf(a[m]);
				location[index5+1+m][2]=location[index5+m][2]+lsom*partZ;
			}

		}
		a+=y;

#if FE_BLADE_DEBUG
		feLog("a %s after\n",c_print(a));
		feLog("w %s\n",c_print(w));
#endif

	}

	velocity[index4]=w[0];
	velocity[index4+1]=w[1];
	velocity[index4+2]=w[2];
	velocity[index4+3]=w[3];
}
