/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_OBLD_DEBUG			FALSE

#include "grass/grass.pmh"

using namespace fe;

void BladeOp::initialize(void)
{
	feLog("BladeOp::initialize\n");

	catalog<Real>("TimeStep")=1.0/60.0;
	catalog<Real>("TimeStep","high")=1e-1;

	catalog<Real>("Tension")=1e4;
	catalog<Real>("Tension","high")=1e5;

	catalog<Real>("Restoration")=0.0;
	catalog<Real>("Restoration","high")=1e5;

	catalog<Real>("Drag")=1e2;
	catalog<Real>("Drag","high")=1e3;

	catalog<Real>("Rest")=1e-2;
	catalog<Real>("Rest","high")=1e0;

	catalog<Real>("Mass")=1e0;
	catalog<Real>("Mass","high")=1e1;

	catalog<Real>("Wind")=0;
	catalog<Real>("Wind","high")=1e3;

	catalog<Real>("Turbulence")=1e2;
	catalog<Real>("Turbulence","high")=1e3;

	catalog<SpatialVector>("Center")=SpatialVector(0.0,0.0,0.0);

	catalog<Real>("Radius")=1e0;
	catalog<Real>("Radius","high")=1e1;

	catalog< sp<Component> >("Input Curves");

	catalog< sp<Component> >("Output Curves");
	catalog<String>("Output Curves","IO")="output";
	catalog<String>("Output Curves","copy")="Input Curves";
	catalog<bool>("Output Curves","recycle")=true;
}

void BladeOp::handle(Record& a_rSignal)
{
#if FE_OBLD_DEBUG
	feLog("BladeOp::handle\n");
#endif

	//* access output curves (copy of input curves)
	m_spOutput=surfaceOutput(a_rSignal);
	if(m_spOutput.isNull())
	{
		feLog("OperatorDodge::handle"
				" output surface not SurfaceAccessibleI\n");
		return;
	}

	m_blade.setTension(catalog<Real>("Tension"));
	m_blade.setRestoration(catalog<Real>("Restoration"));
	m_blade.setDrag(catalog<Real>("Drag"));
	m_blade.setRest(catalog<Real>("Rest"));
	m_blade.setMass(catalog<Real>("Mass"));
	m_blade.setWind(catalog<Real>("Wind"));
	m_blade.setTurbulence(catalog<Real>("Turbulence"));
	m_blade.setCenter(catalog<SpatialVector>("Center"));
	m_blade.setRadius(catalog<Real>("Radius"));

	const Real timeStep=catalog<Real>("TimeStep");
	m_blade.setTimeStep(timeStep);

	static Real t=0.0;	// HACK
	t+=timeStep;
	m_blade.setTime(t);

	m_blade.populate();

	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);

	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	//* inferred reset
	if(pointCount!=m_location.size())
	{
		U32 pointCountV=pointCount*(subCount-1)/subCount;
		U32 pointCountT=pointCount/subCount;
		m_tangent.resize(pointCountT);
		m_location.resize(pointCount);
		m_velocity.resize(pointCountV);

		for(U32 m=0;m<pointCountT;m++)
		{
			const Real angle=sin(0.7*m)+sin(0.3*m);

			set(m_tangent[m],cos(angle),0.0,sin(angle));
//			feLog("%d %.6G %s\n",m,angle,c_print(m_tangent[m]));
		}

		for(U32 m=0;m<pointCount;m++)
		{
			m_location[m]=spAccessorPoint->spatialVector(m,0);
		}

		for(U32 m=0;m<pointCountV;m++)
		{
			m_velocity[m]=0.0;
		}
	}

	setFullRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

#if FE_OBLD_DEBUG
	feLog("BladeOp::handle done\n");
#endif
}

void BladeOp::run(I32 a_id,Range a_range)
{
	sp<SurfaceAccessorI> spAccessorPoint=m_spOutput->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);
	sp<SurfaceAccessorI> spAccessorPrimitive=m_spOutput->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
	const U32 pointCount=spAccessorPoint->count();
	const U32 primitiveCount=spAccessorPrimitive->count();
	const U32 subCount=primitiveCount? pointCount/primitiveCount: 0;

	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

#if FE_OBLD_DEBUG
	feLog("BladeOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	for(U32 index=start;index<=end;index++)
	{
		m_blade.step(index,m_tangent,m_location,m_velocity);

		const U32 startSub=index*subCount;
		const U32 endSub=startSub+subCount;
		for(U32 indexSub=startSub;indexSub<endSub;indexSub++)
		{
			spAccessorPoint->set(indexSub,0,m_location[indexSub]);
		}
	}

#if FE_OBLD_DEBUG
	feLog("BladeOp::run done running thread %d\n",a_id);
#endif
}
