/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_GRASS_DEBUG	FALSE
#define FE_GRASS_PAIRS	FALSE

#include "grass/grass.pmh"

using namespace fe;

void DodgeOp::initialize(void)
{
	//* parameters
	catalog<Real>("Threshold")=0.1;
	catalog<Real>("Falloff")=1.0;
	catalog<Real>("Falloff","high")=10.0;
	catalog<Real>("Falloff","max")=100.0;
	catalog<Real>("Repulsion")=0.1;
	catalog<bool>("Debug Output")=false;

	//* input
	catalog< sp<Component> >("Curves");
	catalog< sp<Component> >("Collider");
}

void DodgeOp::handle(Record& a_rSignal)
{
	const Color blue(0.0,0.0,1.0);
	const Color yellow(1.0,1.0,0.0);

	feLog("DodgeOp::handle\n");

	const Real threshold=catalog<Real>("Threshold");
	const Real falloff=catalog<Real>("Falloff");
	const Real repulsion=catalog<Real>("Repulsion");
	const BWORD debug=catalog<bool>("Debug Output");

	//* access collider
	sp<SurfaceI> spSurfaceCollider;
	sp<SurfaceAccessibleI> spCollider=catalog< sp<Component> >("Collider");
	if(spCollider.isValid())
	{
		sp<SurfaceAccessorI> spColliderAccessor=spCollider->accessor(
				SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);
		U32 colliderPrimitives = spColliderAccessor->count();

		feLog("DodgeOp::handle Collider has %d primitives\n",
				colliderPrimitives);

		spSurfaceCollider=spCollider->surface();
	}

	//* access output curves (copy of input curves)
	sp<SurfaceAccessibleI> spSurfaceAccessibleI=surfaceOutput(a_rSignal);
	if(spSurfaceAccessibleI.isNull())
	{
		feLog("DodgeOp::handle"
				" output surface not spSurfaceAccessibleI\n");
		return;
	}
	sp<SurfaceAccessorI> spAccessor=spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_vertices);

	//* access DrawI interpreter
	sp<DrawI> spDrawI=drawI(a_rSignal);
	if(!spDrawI.isValid())
	{
		feLog("DodgeOp::handle DrawI invalid\n");
		return;
	}
	sp<DrawMode> spDrawMode=spDrawI->drawMode();
	if(!spDrawMode.isValid())
	{
		spDrawMode=new DrawMode();
		spDrawI->setDrawMode(spDrawMode);
	}
	spDrawMode->setGroup("debug");

	//* push curve points out of collider in direction of nearest normal
	const U32 count=spAccessor->count();
	for(U32 index=0;index<count;index++)
	{
		if(!(index%1000))
		{
			feLog("DodgeOp::handle %d/%d\n",index,count);
		}
		BWORD hit=FALSE;
		SpatialVector pivotPoint=spAccessor->spatialVector(index,0);
		SpatialQuaternion pivotQuat;

		const U32 subCount=spAccessor->subCount(index);
		if(!subCount)
		{
			continue;
		}

		SpatialVector lastPoint=spAccessor->spatialVector(index,0);
#if FE_GRASS_PAIRS
		U32 subMax=subCount/2;
#else
		U32 subMax=subCount;
#endif
		//* NOTE skip root point
		for(U32 subIndex=1;subIndex<subMax;subIndex++)
		{
			BWORD changed=FALSE;
			SpatialVector point=spAccessor->spatialVector(index,subIndex);
			const SpatialVector oldPoint=point;

			// HACK quick ignore (TODO as group)
			if(magnitude(point)>10.0)
			{
				break;
			}

			SpatialVector arm1=point-pivotPoint;
			SpatialVector arm2;
			if(hit)
			{
				rotateVector(pivotQuat,arm1,arm2);
				point=pivotPoint+arm2;
			}

#if FE_GRASS_DEBUG
			feLog("\nOperatorDodge::handle %d/%d %d/%d  %s\n",index,count,
					subIndex,subCount,c_print(point));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
#endif

			sp<SurfaceI::ImpactI> spImpact=
					spSurfaceCollider->nearestPoint(point);
			if(spImpact.isValid())
			{
				const SpatialVector intersection=spImpact->intersection();
				const SpatialVector normal=spImpact->normal();
				const SpatialVector contact=intersection+threshold*normal;
				const SpatialVector reach=
						intersection+(threshold+falloff)*normal;
#if FE_GRASS_DEBUG
				feLog("DodgeOp::handle impact %s\n",
						c_print(intersection));
#endif

				const SpatialVector to_contact=contact-point;
				const SpatialVector to_reach=reach-point;
				const Real len2=magnitudeSquared(to_reach);

//				if(len2>16)
//				{
//					break;
//				}

				if(len2>1e-6)
				{
					Real len=sqrt(len2);

					const SpatialVector unitReach=to_reach*(1.0/len);
					Real impactDot=dot(unitReach,normal);
					if(impactDot>0.0)
					{
						const SpatialVector unitSegment=unit(point-lastPoint);
						const SpatialVector projectionSegment=
								unitSegment*dot(unitSegment,normal);
						const SpatialVector sideRepulsion=
								unit(normal-projectionSegment);
						SpatialVector unitRepulsion=sideRepulsion;

						SpatialVector correction=to_contact;
						Real scalarRepulsion=repulsion;

						BWORD draw=FALSE;
						if(falloff>0.0 && len<falloff)
						{
							set(correction);
							const Real ratio=len/falloff;
							scalarRepulsion*=ratio;
							unitRepulsion=unit(sideRepulsion*ratio+
									normal*(1.0-ratio));
						}
						else
						{
							scalarRepulsion-=impactDot*magnitude(correction);
							draw=TRUE;
						}

						if(scalarRepulsion>0.0)
						{
							correction+=scalarRepulsion*unitRepulsion;
						}

//						correction*=pow(impactDot,0.1);
						point+=correction;
						changed=TRUE;

						if(debug && draw)
						{
							SpatialVector line[2];
							line[0]=point;
							line[1]=point+unitRepulsion;

							Color colors[2];
							colors[0]=blue;
							colors[1]=yellow;

							spDrawI->drawLines(line,NULL,2,
									DrawI::e_discrete,false,colors);
						}

						arm1+=pivotPoint;
						pivotPoint=spAccessor->spatialVector(index,subIndex-1);
						arm1-=pivotPoint;
					}
				}
			}
			spAccessor->set(index,subIndex,point);

			arm2=point-pivotPoint;
#if FE_GRASS_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
			feLog("pivotPoint %s\n",c_print(pivotPoint));
			feLog("arm1 %s\n",c_print(arm1));
			feLog("arm2 %s\n",c_print(arm2));
#endif
			if(changed)
			{
				set(pivotQuat,unit(arm1),unit(arm2));
				hit=TRUE;
			}
#if FE_GRASS_DEBUG
			feLog("pivotQuat %s\n",c_print(pivotQuat));
#endif
			lastPoint=oldPoint;
		}
#if FE_GRASS_PAIRS
		for(U32 subIndex=subMax;subIndex<subCount;subIndex++)
		{
			SpatialVector point=
					spAccessor->spatialVector(index,subCount-1-subIndex);
			point[1]+=0.01;
			spAccessor->set(index,subIndex,point);
		}
#endif
	}

	feLog("DodgeOp::handle done\n");
}
