/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

#if FE_SAFE_COUNTED_MUTEX
//* NOTE leak
I32 Memory::ms_mutexCount=0;
boost::recursive_mutex* Memory::ms_pMutex=NULL;
#else
boost::recursive_mutex Memory::ms_mutex;
#endif

#ifdef FE_EXPORT_MEM_VIA_FUNC
FE_DL_EXPORT void *ex_allocate(UWORD byteCount)
{
#if FE_SAFE_COUNTED_MUTEX
	if(Memory::ms_pMutex)
	{
		boost::recursive_mutex::scoped_lock guard(*Memory::ms_pMutex);
	}
	else if(!Memory::ms_mutexCount)
	{
		Memory::ms_mutexCount++;
		Memory::ms_pMutex=new boost::recursive_mutex;
	}
#else
	boost::recursive_mutex::scoped_lock guard(Memory::ms_mutex);
#endif

// NOTE OSX malloc is always 16-byte aligned
#if FE_MEM_ALIGNMENT && FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64 && FE_OS!=FE_OSX
	void* pBuffer;
	int fail=posix_memalign(&pBuffer,FE_MEM_ALIGNMENT,byteCount);
//	printf("ex_allocate(%d) %p %d\n",int(byteCount),pBuffer,fail);
//	FEASSERT(pBuffer<(void*)0x200000000000);
#if FE_CHECK_HEAP
	if(!fail && (!gs_pHeapBase || gs_pHeapBase>pBuffer))
	{
		gs_pHeapBase=pBuffer;
	}
#elif FE_CHECK_NEW
	gs_pHeapBase=(void*)0x1;
#endif
	return fail? NULL: pBuffer;
#elif FE_CODEGEN<=FE_DEBUG && (FE_OS==FE_WIN32 || FE_OS==FE_WIN64)
	void* pBuffer=malloc(byteCount);
#if FE_CHECK_HEAP
	if(!gs_pHeapBase || gs_pHeapBase>pBuffer)
	{
		gs_pHeapBase=pBuffer;
	}
#elif FE_CHECK_NEW
	gs_pHeapBase=(void*)0x1;
#endif
	return pBuffer;
#else
	return malloc(byteCount);
#endif
}

FE_DL_EXPORT void *ex_reallocate(void *pMemory, UWORD byteCount)
{
#if FE_SAFE_COUNTED_MUTEX
	if(Memory::ms_pMutex)
	{
		boost::recursive_mutex::scoped_lock guard(*Memory::ms_pMutex);
	}
	else if(!Memory::ms_mutexCount)
	{
		Memory::ms_mutexCount++;
		Memory::ms_pMutex=new boost::recursive_mutex;
	}
#else
	boost::recursive_mutex::scoped_lock guard(Memory::ms_mutex);
#endif

#if FE_MEM_ALIGNMENT && FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64 && FE_OS!=FE_OSX
	void *pNew=realloc(pMemory, byteCount);
	if(pNew==pMemory)
	{
		return pNew;
	}

	void* pBuffer;
	int fail=posix_memalign(&pBuffer,FE_MEM_ALIGNMENT,byteCount);
//	printf("ex_reallocate(%d) %p %d\n",int(byteCount),pBuffer,fail);
	if(fail)
	{
		free(pNew);
		return NULL;
	}
	memcpy(pBuffer,pNew,byteCount);
	free(pNew);

	return pBuffer;
#else
	return realloc(pMemory, byteCount);
#endif
}

FE_DL_EXPORT void ex_deallocate(void *pMemory)
{
#if FE_SAFE_COUNTED_MUTEX
	if(Memory::ms_pMutex)
	{
		boost::recursive_mutex::scoped_lock guard(*Memory::ms_pMutex);
	}
	else if(!Memory::ms_mutexCount)
	{
		Memory::ms_mutexCount++;
		Memory::ms_pMutex=new boost::recursive_mutex;
	}
#else
	boost::recursive_mutex::scoped_lock guard(Memory::ms_mutex);
#endif

	free(pMemory);
}

#endif

#ifdef FE_EXPORT_MEMORY
#ifdef FE_EXPORT_MEM_VIA_FUNC
allocateFunction	*gs_pAllocateFunction	= ex_allocate;
deallocateFunction	*gs_pDeallocateFunction	= ex_deallocate;
reallocateFunction	*gs_pReallocateFunction	= ex_reallocate;
void*				gs_pHeapBase			= NULL;
I32					gs_newCheck				= 0;
#else
#if FE_COMPILER==FE_DMC
allocateFunction	*gs_pAllocateFunction	= malloc;
deallocateFunction	*gs_pDeallocateFunction	= free;
reallocateFunction	*gs_pReallocateFunction	= realloc;
void*				gs_pHeapBase			= NULL;
I32					gs_newCheck				= 0;
#else
FE_DL_EXPORT allocateFunction	*gs_pAllocateFunction	= malloc;
FE_DL_EXPORT deallocateFunction	*gs_pDeallocateFunction	= free;
FE_DL_EXPORT reallocateFunction	*gs_pReallocateFunction	= realloc;
FE_DL_EXPORT void*				gs_pHeapBase			= NULL;
FE_DL_EXPORT I32				gs_newCheck				= 0;
#endif
#endif
#endif

} /* namespace */
